package lab2;

import  lab2.Exceptions.*;

/**
 * Classe HashSet:
 *
 * Semplice gestione di un insieme non ordinato mediante tabelle hash
 * con liste concatenate.  Possibilit&agrave; di avere un semplice
 * iteratore (cursore), con riferimento alle classi di Java
 * Collections API disponibili in <A
 * HREF="http://java.sun.com/products/jdk/1.2/docs/guide/collections/reference.html">
 * <TT>java.util.*</TT> (JDK 1.2)</A> [HashSet]. Gli oggetti vengono
 * inseriti in posizioni della tabella che dipendono dalla funzione
 * hash adottata. Il metodo <code>remove</code> <I>con</I> argomento
 * si riferisce all'interfaccia <code>List</code>; il metodo
 * <code>remove</code> <I>senza</I> argomento si riferisce
 * all'interfaccia <code>Iterator</code>. Quest'ultimo metodo evita
 * modifiche concorrenti che possano interferire con altre istanze di
 * iteratori mediante contatori di modifiche.
 *
 * @see lab2.List
 * @see lab2.ListNode
 * @see lab2.LinkedList
 * @see lab2.Set
 * @version 2.0 Febbraio 2001
 * @author A. Brogi      (<A HREF="mailto:brogi@di.unipi.it">brogi@di.unipi.it</A>)
 * @author A. Corradini  (<A HREF="mailto:andrea@di.unipi.it">andrea@di.unipi.it</A>)
 * @author F. Gadducci   (<A HREF="mailto:gadducci@di.unipi.it">gadducci@di.unipi.it</A>)
 * @author R. Grossi     (<A HREF="mailto:grossi@di.unipi.it">grossi@di.unipi.it</A>)
 * @author C. Montangero (<A HREF="mailto:monta@di.unipi.it">monta@di.unipi.it</A>)
 * @author F. Romani     (<A HREF="mailto:romani@di.unipi.it">romani@di.unipi.it</A>)
 * @author L. Semini     (<A HREF="mailto:semini@di.unipi.it">semini@di.unipi.it</A>)
 */
public class HashSet implements Set {
    protected static final int DEFAULT_SIZE = 131;
    protected List [ ] theLists;  // Tabella hash con liste concatenate
    protected int modCount;    // Numero di modifiche effettuate nell'insieme

    /**
     * Costruttore di tabella hash di dimensione prefissata
     * (<tt>DEFAULT_SIZE</tt>).  
     */
    public HashSet( ) {
        this( DEFAULT_SIZE );
    }

   /**
     * Costruttore di tabella hash di dimensione a scelta. L'intero
     * passato viene trasformato nel successivo numero primo, che
     * diventa la dimensione della tabella.  In alternativa, si
     * pu&ograve; passare direttamente un numero primo.  
     */
    public HashSet( int tableSize ) {
        theLists = new LinkedList[ nextPrime( tableSize ) ];

        for ( int i = 0; i < theLists.length; i++ )
            theLists[ i ] = new LinkedList( );
	modCount = 0;
    }

    /**
     * Verifica che l'insieme sia logicamente vuoto.
     * @return  <code>true</code> se l'insieme &egrave; vuoto;
     *          <code>false</code> altrimenti.
     */
    public boolean isEmpty( ) {
	// Esercizio: usare un contatore globale size
        for ( int i = 0; i < theLists.length; i++ ) {
            if ( !theLists[ i ].isEmpty( ) )
                return false;
        }
        return true;
    }

    /**
     * Rende l'insieme vuoto.
     */
    public void clear( ) {
	modCount++;
        for ( int i = 0; i < theLists.length; i++ )
            theLists[ i ].clear( );
    }

    // Metodo interno per determinare la lista di un dato oggetto (non null)
    protected List listOf( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
        int index = x.hashCode( );  
                // Potrebbe essere negativo a causa di un overflow

        index = index % theLists.length;
        if ( index < 0 )
            index += theLists.length;
        return theLists[ index ];
    }

    /**
     * Verifica se l'insieme contiene l'oggetto specificato, usando il
     * metodo standard <tt>equals</tt> (l'oggetto pu&ograve; occorrere
     * al pi&ugrave; una volta).
     * @param x l'oggetto da cercare nell'insieme.
     * @return  <code>true</code> se l'insieme contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean contains( Object x ) {
        List whichList = listOf( x );

        return whichList.contains( x );
    }

    /**
     * Aggiunge un oggetto all'insieme. Non sono ammessi oggetti null
     * e/o duplicati.
     * @param x l'oggetto da inserire, <em>diverso</em> da null.
     * @return <code>true</code> se l'oggetto non apparteneva
     * gi&agrave; all'insieme; <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean add( Object x ) {
        List whichList = listOf( x );
        boolean isNew = !whichList.contains( x );

        if ( isNew ) {
            whichList.add( x );
	    modCount++;
	}
        return isNew;
    }

    /**
     * Elimina un oggetto dall'insieme, che rimane inalterato se non
     * contiene l'oggetto. Il metodo standard <tt>equals</tt> viene
     * impiegato per trovare tale oggetto.

     * @param x  l'oggetto da eliminare, se presente e <em>diverso</em> 
     * da null. 
     * @return  <code>true</code> se l'oggetto appartiene all'insieme;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean remove( Object x ) {
        List whichList = listOf( x );
	boolean isModified = whichList.remove( x );

	if ( isModified )
	    modCount++;
        return isModified; 
    }

    // Metodo interno per trovare il minimo numero primo >= n
    protected static int nextPrime( int n ) {
	if ( n > 1 ) {
	    // Esiste sempre un primo tra n e 2n [Bertrand, Chebyshev]
	    boolean [] P = primes( 2*n+1 ); 
	    for ( int i = n; i < P.length; i++ )
		if ( P[i] )
		    return i;
	}
	return 2;
    }

    /* Metodo interno per la primalita' (crivello di Eratostene):
       rende un vettore di booleani che vale true nelle posizioni
       corrispondenti ai numeri primi */
    protected static boolean[ ] primes( int n ) {
	boolean [] primeList = new boolean[n];
	
	primeList[0] = primeList[1] = false;
	for ( int i = 2; i < n; i++ )
	    primeList[i] = true;

	int i = 1;
	while ( i*i < n ) {
	    // Trova il prossimo primo a partire da i
	    while ( !primeList[++i] ) ;
	    // Adesso i e' uguale a tale primo
	    int k = i*i;
	    while (k < n) {
		primeList[k] = false;
		k += i;
	    }
	} 
	return primeList;
    }

    /**
     * Restituisce un iteratore per scandire gli oggetti dell'insieme
     * secondo la posizione assegnata a loro dalla funzione hash. Se
     * l'insieme viene modificato con metodi diversi da quelli
     * dell'iteratore, viene generata una eccezione
     * <tt>ConcurrentModificationException</tt>. In questo modo,
     * l'iterazione termina per evitare comportamenti difficili da
     * prevedere.
     * @return una classe che realizza <code>Iterator</code> per
     * scandire gli oggetti dell'insieme.
     * @exception lab2.Exceptions.ConcurrentModificationException se
     * avviene una modifica concorrente mediante altri metodi.  
     * @see java.util.Iterator 
     */
    public Iterator iterator( ) {
        return new ProtectedItr( );
    }

    protected class ProtectedItr implements Iterator {
	/*
	  INVARIANTE: Il cursore e' rappresentato dall'iteratore
	  listCursor sulla lista corrente e dall'indice tableCursor
	  della prossima lista non vuota (cioe' tableCursor <
	  theLists.length, altrimenti non c'e' una tale lista).

	  NOTA: l'invariante e' motivata dal fatto che uno puo'
	  lasciare la lista corrente solo prima di invocare il metodo
	  next, per via del metodo remove. 
	  Esercizio: Realizzare l'iteratore senza remove.
	*/
        protected int tableCursor;  
            // Il cursore sulla tabella di liste concatenate
        protected Iterator listCursor;  
            // Il cursore sulla lista corrente nella tabella

        protected int expectedModCount; 
            // Per accorgersi di eventuali modifiche concorrenti

	/* 
	   Metodo interno per trovare la prima lista non vuota a
	   partire da tableCursor (escluso) 
	*/
	void nextList( ) {
	    do {
		++tableCursor;
	    } while ( ( tableCursor < theLists.length )
		      && theLists[tableCursor].isEmpty( ) );
	}

	/**
	 * Costruttore
	 */
	ProtectedItr( ) {
	    tableCursor = -1;
	    nextList( );
	    if ( tableCursor < theLists.length ) {
		listCursor = theLists[tableCursor].iterator( );
		nextList( );
	    } else {
		listCursor = null;
	    }
	    expectedModCount = modCount; 
                // Per il controllo delle modifiche concorrenti
	}

	/**
          * Indica se ci sono ulteriori oggetti da esaminare
          * nell'iterazione corrente.
          * @return <tt>true</tt> se ci sono ulteriori oggetti;
          *         <tt>false</tt> altrimenti.  
	  * @exception ConcurrentModificationException se avviene
	  *            una modifica concorrente degli stessi oggetti
	  *            mediante qualche altro metodo.
          */
         public boolean hasNext( ) {
             if ( modCount != expectedModCount )  // Modifica concorrente
                 throw new ConcurrentModificationException( );
	     if ( listCursor == null )
		 return false;
	     if ( listCursor.hasNext( ) )
		 return true;
	     return ( tableCursor < theLists.length );
         }
    
        /**
         * Restituisce l'oggetto successivo da esaminare
         * nell'iterazione. 
         * @return il prossimo oggetto da esaminare.
         * @exception NoSuchElementException se non ci sono altri
         *            oggetti da esaminare.  
         * @exception ConcurrentModificationException se avviene
         *            una modifica concorrente degli stessi oggetti
         *            mediante qualche altro metodo.
	 */
        public Object next( ) {
             if ( !hasNext( ) )                   // Tabella vuota
                 throw new NoSuchElementException( );
	     if ( listCursor.hasNext( ) )
		 return listCursor.next( );
	     // tableCursor < theLists.length
	     listCursor = theLists[tableCursor].iterator( );
	     nextList( );
	     return listCursor.next( );
        }
    
        /**
         * Cancella l'oggetto restituito dall'ultima chiamata al
         * metodo <code>next</code>.  Il metodo pu&ograve; essere
         * invocato solo una volta per ciascuna chiamata a
         * <code>next</code>. 
         * @exception NoSuchElementException se non ci sono altri
         *            oggetti da esaminare.  
         * @exception ConcurrentModificationException se avviene
         *            una modifica concorrente degli stessi oggetti
         *            mediante qualche altro metodo.
         */
        public void remove( ) {
	    if ( modCount++ != expectedModCount++ )  // Modifica concorrente
		throw new ConcurrentModificationException( );
             if ( listCursor == null )
                 throw new NoSuchElementException( );
	    listCursor.remove( );
        }
    }
}

