package lab2;

import lab2.Exceptions.*;

/**
 * Classe LinkedList:
 *
 * Semplice gestione di una lista unidirezionale con la
 * possibilit&agrave; di avere un iteratore (cursore). Si vedano le
 * classi di Java Collections API disponibili in <A HREF=
 * "http://java.sun.com/products/jdk/1.2/docs/guide/collections/reference.html">
 * <TT>java.util.*</TT> (JDK 1.2)</A> [LinkedList]. Si noti che
 * esistono altre possibili realizzazioni di lista con metodi diversi
 * da quelli presentati e usando altri tipi di iteratori. Il metodo
 * <code>remove</code> <I>con</I> argomento si riferisce
 * all'interfaccia <code>List</code>; il metodo <code>remove</code>
 * <I>senza</I> argomento si riferisce all'interfaccia
 * <code>Iterator</code>. Quest'ultimo metodo evita modifiche
 * concorrenti che possano interferire con altre istanze di iteratori
 * mediante contatori di modifiche.
 *
 * @see lab2.List
 * @see lab2.ListNode
 * @version 2.0 Febbraio 2001
 * @author A. Brogi      (<A HREF="mailto:brogi@di.unipi.it">brogi@di.unipi.it</A>)
 * @author A. Corradini  (<A HREF="mailto:andrea@di.unipi.it">andrea@di.unipi.it</A>)
 * @author F. Gadducci   (<A HREF="mailto:gadducci@di.unipi.it">gadducci@di.unipi.it</A>)
 * @author R. Grossi     (<A HREF="mailto:grossi@di.unipi.it">grossi@di.unipi.it</A>)
 * @author C. Montangero (<A HREF="mailto:monta@di.unipi.it">monta@di.unipi.it</A>)
 * @author F. Romani     (<A HREF="mailto:romani@di.unipi.it">romani@di.unipi.it</A>)
 * @author L. Semini     (<A HREF="mailto:semini@di.unipi.it">semini@di.unipi.it</A>)
 */
public class LinkedList implements List {
    protected ListNode first;  // inizio lista
    protected ListNode last;   // fine lista
    protected int modCount;    // Numero di modifiche effettuate nell'insieme

    /**
     * Costruttore di lista vuota
     */
    public LinkedList( ) {
        first = last = null;
        modCount = 0;
    }

    /**
     * Verifica che la lista sia logicamente vuota.
     * @return  <code>true</code> se la lista &egrave; vuota;
     *          <code>false</code> altrimenti.
     */
    public boolean isEmpty( ) {
        return first == null;
    }

    /**
     * Svuota la lista.
     */
    public void clear( ) {
	modCount++;
	first = last = null;
    }

    /**
     * Verifica se la lista contiene l'oggetto specificato, usando il
     * metodo standard <tt>equals</tt> (l'oggetto pu&ograve; occorrere
     * pi&ugrave; volte).
     * @param x l'oggetto da cercare nella lista.
     * @return  <code>true</code> se la lista contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean contains( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
        for ( ListNode h = first; h != null; h = h.next ) {
            if ( x.equals( h.element ) ) 
                return true;
        }
        return false;
    }

    /**
     * Inserisce un oggetto in fondo alla lista. Non sono ammessi 
     * oggetti null.
     * @param x l'oggetto da inserire, <em>diverso</em> da null.
     * @return  <code>true</code> come richiesto da Java Collection API.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean add( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
        modCount++;
        if ( isEmpty( ) ) {   // Primo elemento
            first = last = new ListNode( x );
        } else {              // Elementi successivi
            last.next = new ListNode( x );
            last = last.next;
        }
        return true;
    }

    /**
     * Cancella la prima occorrenza dell'oggetto a partire dall'inizio
     * della lista, che rimane inalterata se non contiene l'oggetto. 
     * Il metodo standard <tt>equals</tt> viene impiegato per trovare 
     * la prima occorrenza.
     * @param x  l'oggetto da cancellare, se presente e <em>diverso</em>
     * da null. 
     * @return  <code>true</code> se la lista contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.  
     */
    public boolean remove( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
	if ( isEmpty( )  )
	    return false;
	if ( x.equals( first.element ) ) { // Primo elemento
	    if ( first == last )      // ... e ultimo elemento
		last = null;
	    first = first.next;
	    modCount++;
	    return true;
	}
	// Elementi successivi
	for ( ListNode h = first; h.next != null; h = h.next ) {
	    if ( x.equals( h.next.element ) ) {
		if ( h.next == last )      // Ultimo elemento
		    last = h;
		h.next = h.next.next;
		modCount++;
		return true;
	    }
	}
        return false;
    }
 
    /**
     * Restituisce un iteratore sulla lista per scandire gli oggetti
     * in sequenza a partire dall'inizio della lista. Se la lista
     * &egrave; modificata con metodi diversi da quelli
     * dell'iteratore, viene generata una eccezione
     * <tt>ConcurrentModificationException</tt>. In tal caso,
     * l'iterazione termina per evitare comportamenti difficili
     * da prevedere.
     * @return una classe che realizza <code>Iterator</code> per
     * scandire gli oggetti in sequenza dall'inizio.
     * @exception lab2.Exceptions.ConcurrentModificationException se
     * avviene una modifica concorrente mediante altri metodi.
     */
    public Iterator iterator( ) {
        return new ProtectedItr( );
    }

    protected class ProtectedItr implements Iterator {
	/*
	  INVARIANTE: cursor individua il nodo che verra` restituito
	  dalla prossima chiamata di next; lo fa in generale
	  mantenendo il puntatore al nodo che precede quello da
	  restituire; per il nodo (virtuale) che precede l'eventuale
	  primo nodo della lista (puntato da first) si usa null. Per
	  quanto riguarda precursor, esso coincide con cursor, tranne
	  quando e` possibile eliminare un elemento, nel qual caso
	  precede cursor, cioe` individua il nodo precedente.

	  NOTA: In particolare, quindi, cursor e precursor coincidono
	  dopo l'invocazione del costruttore dell'iteratore dato che
	  non e` possibile rimuovere(e sono entrambi null), e dopo
	  l'invocazione del metodo remove; invece precursor precede
	  cursor dopo l'invocazione di next (e quindi vale null dopo
	  l'esame del primo della lista, in base alla stessa
	  convenzione usata per cursor).

	  NOTA: La scelta di stare un passo indietro permette di
	  eliminare i nodi, ma comporta che il primo nodo sia trattato
	  come caso a parte.
	*/
        protected ListNode cursor;  
            // Il cursore implicitamente gestito nell'iteratore
        protected ListNode precursor;  
            // Precede oppure e' uguale al cursore

        protected int expectedModCount; 
            // Per accorgersi di eventuali modifiche concorrenti

         /**
          * Costruttore
          */
         ProtectedItr( ) {
             precursor = cursor = null;
             expectedModCount = modCount; 
                 // Per il controllo delle modifiche concorrenti
         }

         /**
          * Indica se ci sono ulteriori nodi da esaminare
          * nell'iterazione corrente.
          * @return <tt>true</tt> se ci sono ulteriori oggetti;
          *         <tt>false</tt> altrimenti.  
	  * @exception ConcurrentModificationException se avviene
	  *            una modifica concorrente degli stessi oggetti
	  *            mediante qualche altro metodo.
          */
         public boolean hasNext( ) {
             if ( modCount != expectedModCount )  // Modifica concorrente
                 throw new ConcurrentModificationException( );
             return ( !isEmpty( ) && 
		      ( (cursor == null) || (cursor.next != null) ) );
         }
    
        /**
         * Restituisce l'oggetto successivo da esaminare
         * nell'iterazione.  
         * @return il prossimo oggetto da esaminare.
         * @exception NoSuchElementException se non ci sono altri
         *            oggetti da esaminare.  
         * @exception ConcurrentModificationException se avviene
         *            una modifica concorrente degli stessi oggetti
         *            mediante qualche altro metodo.
         */
        public Object next( ) {
             if ( !hasNext( ) )                   // Fuori lista
                 throw new NoSuchElementException( );
             precursor = cursor;
             cursor = (cursor == null) ? first : cursor.next;
             return cursor.element;
        }
    
        /**
         * Cancella l'oggetto restituito dall'ultima chiamata al
         * metodo <code>next</code>.  Il metodo pu&ograve; essere
         * invocato solo una volta per ciascuna chiamata a
         * <code>next</code>. 
         * @exception ConcurrentModificationException se avviene
         * una modifica concorrente degli stessi oggetti
         * mediante qualche altro metodo.
         * @exception IllegalStateException se nessuna chiamata al
         * metodo <tt>next</tt> &egrave; stata precedentemente
         * effettuata, oppure una chiamata a <tt>remove</tt> &egrave;
         * stata gi&agrave; eseguita in corrispondenza dello stessa
         * chiamata a <tt>next</tt>.  
         */
        public void remove( ) {
	    if ( modCount++ != expectedModCount++ )  // Modifica concorrente
		throw new ConcurrentModificationException( );
	    if ( cursor == precursor ) // Chiamata non legale
		throw new IllegalStateException( );
	    if ( precursor == null ) { // Cursor punta al primo elemento
		if ( first == last )   // ... e ultimo elemento
		    last = null;
		first = first.next;
	    } else {                   // Cursor punta ai nodi successivi
		if ( cursor == last )  // Ultimo elemento
		    last = precursor;
		precursor.next = cursor.next;
	    }
	    cursor = precursor;
        }    
    }
}



