package lab2;

import  lab2.Exceptions.*;

/**
 * Classe TreeSet:
 *
 * Gestione di insiemi <em>ordinati</em> mediante alberi binari di
 * ricerca <I>non</I> bilanciati. La differenza con la
 * classe <tt>HashSet</tt> &egrave; quindi nella possibilit&agrave; di
 * mantenere un ordine totale, come realizzato nelle classi di Java
 * Collections API disponibili in <A
 * HREF="http://java.sun.com/products/jdk/1.2/docs/guide/collections/reference.html">
 * <TT>java.util.*</TT> (JDK 1.2)</A> [TreeSet]. Gli oggetti
 * dell'insieme devono fornire il metodo <tt>compareTo</tt>
 * dell'interfaccia standard <tt>Comparable</tt> per mantenere le
 * chiavi ordinate.  L'eccezione <tt>IllegalArgumentException</tt> viene
 * generata se gli oggetti non sono tutti
 * <tt>Comparable</tt>. &Egrave; possibile, infine, creare un
 * iteratore <code>Iterator</code> che scandisce gli oggetti in
 * maniera crescente secondo l'ordine specificato. Il metodo
 * <code>remove</code> <I>con</I> argomento si riferisce
 * all'interfaccia <code>Set</code>; il metodo <code>remove</code>
 * <I>senza</I> argomento si riferisce all'interfaccia
 * <code>Iterator</code>. Quest'ultimo metodo evita modifiche
 * concorrenti che possano interferire con altre istanze di iteratori
 * mediante contatori di modifiche.
 *
 * @see lab2.Set
 * @see lab2.BinNode
 * @version 2.0 Febbraio 2001 
 * @author A. Brogi      (<A HREF="mailto:brogi@di.unipi.it">brogi@di.unipi.it</A>)
 * @author A. Corradini  (<A HREF="mailto:andrea@di.unipi.it">andrea@di.unipi.it</A>)
 * @author F. Gadducci   (<A HREF="mailto:gadducci@di.unipi.it">gadducci@di.unipi.it</A>)
 * @author R. Grossi     (<A HREF="mailto:grossi@di.unipi.it">grossi@di.unipi.it</A>)
 * @author C. Montangero (<A HREF="mailto:monta@di.unipi.it">monta@di.unipi.it</A>)
 * @author F. Romani     (<A HREF="mailto:romani@di.unipi.it">romani@di.unipi.it</A>)
 * @author L. Semini     (<A HREF="mailto:semini@di.unipi.it">semini@di.unipi.it</A>)
 */

public class TreeSet implements Set {
    protected BinNode root;        // Radice (con padre nullo)
    protected int modCount;        // Numero di modifiche effettuate

    /**
     * Costruttore dell'albero 
     */
    public TreeSet( ) {
        root = null;
        modCount = 0;
    }

    /**
     * Verifica che l'insieme sia logicamente vuoto.
     * @return  <code>true</code> se l'insieme &egrave; vuoto;
     *          <code>false</code> altrimenti.
     */
    public boolean isEmpty( ) {
        return (root == null);
    }

    /**
     * Rende l'insieme vuoto.
     */
    public void clear( ) {
        modCount++;
        root = null;
    }

    /**
     * Verifica se l'insieme contiene l'oggetto specificato.
     * @param x l'oggetto da cercare nell'insieme.
     * @return  <code>true</code> se l'insieme contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code> oppure non &egrave;
     *            <code>Comparable</code>.
     */
    public boolean contains ( Object x ) {
	if ( (x == null) || !(x instanceof Comparable) )
	    throw new IllegalArgumentException( );
	return ( containsHelper( (Comparable) x ) != null );
    }

    /*
      Metodo interno che restituisce il nodo contenente l'oggetto x se
      questo appartiene all'albero; restituisce null altrimenti.  
    */
    protected BinNode containsHelper ( Comparable cx ) {
	BinNode t = root;

	while ( t != null ) {
	    if ( cx.compareTo( t.key ) < 0 ) 
		t = t.left;
	    else if ( cx.compareTo( t.key ) > 0 ) 
		t = t.right;
	    else
		return t;
	}
	return null;
    }

    /**
     * Aggiunge un oggetto all'insieme. Non sono ammessi 
     * oggetti null e/o duplicati.
     * @param x l'oggetto da inserire, <em>diverso</em> da null.
     * @return  <code>true</code> se l'oggetto x non appartiene
     *           all'insieme; <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code> oppure non &egrave;
     *            <code>Comparable</code>.
     */
    public boolean add( Object x ) {
	if ( (x == null) || !(x instanceof Comparable) )
	    throw new IllegalArgumentException( );
	if (isEmpty()) {     // Crea l'albero
	    root = new BinNode (x);
	    modCount++;
	    return true;
	}
	return ( addHelper( (Comparable)x, root ) );
    }

    /*
      Metodo interno che inserisce l'oggetto x nell'albero non vuoto
      t.  Restituisce true se x non appartiene gia' all'albero;
      restituisce false altrimenti. 
    */
    protected boolean addHelper ( Comparable cx, BinNode t ) {
	if ( cx.compareTo( t.key ) < 0 ) {
	    if (t.left == null){
		// Inserisce x come figlio sinistro di t
		t.left = new BinNode ( cx, t );
		modCount++;
		return true;
	    }
	    // Altrimenti la ricerca continua nel sottoalbero sinistro
	    return addHelper( cx, t.left );
	}
	if ( cx.compareTo( t.key ) > 0 ) {
	    if (t.right == null) {
		// Inserisce x come figlio destro di t
		t.right = new BinNode ( cx, t );
		modCount++;
		return true;
	    }
	    // Altrimenti la ricerca continua nel sottoalbero destro
	    return addHelper( cx, t.right );		
	}
	return false;
    }

    /**
     * Elimina l'oggetto dall'insieme, che rimane inalterato se non 
     * contiene l'oggetto. 
     * @param x  l'oggetto da eliminare, se presente e <em>diverso</em> 
     * da null. 
     * @return  <code>true</code> se l'oggetto appartiene all'insieme;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code> oppure non &egrave;
     *            <code>Comparable</code>.
     */
    public boolean remove( Object x ) {
	if ( (x == null) || !(x instanceof Comparable) )
	    throw new IllegalArgumentException( );
	BinNode z = containsHelper( (Comparable)x );
	if ( z != null ) {
	    // x e' contenuto nel nodo z dell'albero
	    removeHelper( z );
	    modCount++;
	    return true;
	}
	return false;
    }

    /*
      Metodo interno per cancellare il nodo z dall'albero
    */
    protected void removeHelper ( BinNode z ) {
	BinNode y = z; // Nodo da eliminare
	if ( (z.left != null) && (z.right != null) ) {
	    // z ha due figli: va eliminato il successore di z
	    y = successor( z );
	    z.key = y.key;
	}
	// In ogni caso, y ha al piu' un figlio x
	BinNode x = ( y.left !=null ) ? y.left : y.right;
	// Per rimuovere y, si aggiustano i riferimenti a padri e figli
	if (x != null)
	    x.parent = y.parent;
	if (y.parent == null)
	    root = x;    // Radice
	else 
	    if (y == y.parent.left)
		y.parent.left = x;
	    else
		y.parent.right = x;
    }

    /*
      Metodo interno per trovare il successore di un nodo
      dell'albero. Se il nodo contiene la chiave x, allora il
      successore e' il nodo contenente la piu' piccola tra le chiavi
      piu' grandi di x. Se tale nodo non esiste, restituisce null.
    */
    protected BinNode successor( BinNode t ) {
        if (t == null)
            return null;
	if (t.right != null)
            return min( t.right );  // Minimo del sottoalbero destro
	/* Altrimenti: antenato piu' prossimo a t il cui figlio
	   sinistro sia anch'esso antenato di t */
	BinNode y = t.parent;
	while ( (y != null) && (t == y.right) ) {
	    t = y; 
	    y = y.parent; 
	}
	return y;  
    }

    /*
      Metodo interno per trovare il predecessore di un nodo
      dell'albero. Simmetrico al successore. Se tale nodo non esiste,
      restituisce null.  
    */
    protected BinNode predecessor( BinNode t ) {
        if (t == null)
            return null;
	if (t.left != null)
            return max( t.left );  // Massimo del sottoalbero sinistro
	/* Altrimenti: antenato piu' prossimo a t il cui figlio destro
	   sia anch'esso antenato di t */
	BinNode y = t.parent;
	while ( (y != null) && (t == y.left) ) {
	    t = y; 
	    y = y.parent; 
	}
	return y; 
    }

    /**
       Restituisce la chiave pi&ugrave; piccola dell'albero.
       @return <code>null</code> se l'albero &egrave; vuoto.  
    */
    public BinNode min( ) {
	return min( root );
    }

    protected BinNode min( BinNode t ) {
	if ( t == null ) 
	    return null;
	while ( t.left != null )
	    t = t.left;
	return t;
    }

    /**
       Restituisce la chiave pi&ugrave; grande dell'albero.
       @return <code>null</code> se l'albero &egrave; vuoto.  
    */
    public BinNode max( ) {
	return max( root );
    }

    protected BinNode max( BinNode t ) {
	if ( t == null ) 
	    return null;
	while ( t.right != null )
	    t = t.right;
	return t;
    }

    /**
     * Restituisce un iteratore per scandire gli oggetti dell'insieme
     * in ordine crescente secondo l'ordine indotto da
     * <tt>compareTo</tt>. Se l'insieme viene modificato con metodi
     * diversi da quelli definiti dall'iteratore, viene generata una
     * eccezione <tt>ConcurrentModificationException</tt>. In questo
     * modo, l'iterazione termina per evitare comportamenti difficili
     * da prevedere.
     * @return una classe che realizza <code>Iterator</code> per
     * scandire gli oggetti dell'insieme.
     * @exception ConcurrentModificationException se avviene una
     * modifica concorrente mediante altri metodi.
     * @see java.util.Iterator 
     */
    public Iterator iterator( ) {
        return new ProtectedItr( );
    }

    protected class ProtectedItr implements Iterator {
	/*
	  INVARIANTE: Identica a quella di Iterator in LinkedList,
	  eccetto che qui viene usato successor(cursor) invece che
	  cursor.next, dove cursor individua il nodo dell'albero che
	  verra` restituito dalla prossima chiamata di next; lo fa in
	  generale mantenendo il puntatore al nodo che precede quello
	  da restituire; per il nodo (virtuale) che precede
	  l'eventuale nodo minimo dell'albero si usa null. Per quanto
	  riguarda precursor, esso coincide con cursor, tranne quando
	  e` possibile eliminare un elemento, nel qual caso precede
	  cursor, cioe` individua il nodo precedente.
	*/
	protected BinNode cursor; 
	    // Il cursore implicitamente gestito nell'iteratore
	protected BinNode precursor;
	    // Precede oppure e' uguale al cursore
	
	protected int expectedModCount; 
	  // Per accorgersi di eventuali modifiche concorrenti

        /**
         * Costruttore 
	 */
        ProtectedItr( ) {
            precursor = cursor = null;
            expectedModCount = modCount; 
                    // Per il controllo delle modifiche concorrenti
        }

        /**
         * Indica se ci sono ulteriori elementi da esaminare
         * nell'iterazione corrente.
         * @return <tt>true</tt> se ci sono ulteriori elementi;
         *         <tt>false</tt> altrimenti.  
	 * @exception ConcurrentModificationException se avviene
	 *            una modifica concorrente degli stessi oggetti
	 *            mediante qualche altro metodo.
         */
        public boolean hasNext( ) {
	    if ( modCount != expectedModCount )  // Modifica concorrente
		throw new ConcurrentModificationException( );
            return ( !isEmpty( ) && 
		     ( (cursor == null) || (successor( cursor ) != null) ) );
        }
    
        /**
         * Restituisce l'oggetto successivo da esaminare
         * nell'iterazione.
         * @return il prossimo oggetto da esaminare.
         * @exception NoSuchElementException se non ci sono altri
         *            oggetti da esaminare.  
         * @exception ConcurrentModificationException se avviene
         *            una modifica concorrente degli stessi oggetti
         *            mediante qualche altro metodo.
         */
        public Object next( ) {
            if ( !hasNext( ) )                   // Esaminati tutti i nodi
                throw new NoSuchElementException( );
	    precursor = cursor;
	    // Il primo elemento e' min(root), non necessariamente root
            cursor = (cursor == null) ? min( root ) : successor( cursor );
            return cursor.key;
        }
    
        /**
         * Cancella l'oggetto restituito dall'ultima chiamata al
         * metodo <code>next</code>.  Il metodo pu&ograve; essere
         * invocato solo una volta per ciascuna chiamata a
         * <code>next</code>. 
         * @exception ConcurrentModificationException se avviene
         * una modifica concorrente degli stessi oggetti
         * mediante qualche altro metodo.
         * @exception IllegalStateException se nessuna chiamata al
         * metodo <tt>next</tt> &egrave; stata precedentemente
         * effettuata, oppure una chiamata a <tt>remove</tt> &egrave;
         * stata gi&agrave; eseguita in corrispondenza dello stessa
         * chiamata a <tt>next</tt>.  
         */
        public void remove( ) {
            if ( modCount++ != expectedModCount++ )  // Modifica concorrente
                throw new ConcurrentModificationException( );
            if ( cursor == precursor )               // Chiamata non legale
                throw new IllegalStateException( );
	    removeHelper( cursor );
            cursor = precursor;
        }    
    }
}
