package lab2;

import  lab2.Exceptions.*;
import  java.util.Vector;

/**
 * Classe VectorList:
 *
 * Realizzazione della Lista mediante Vector, utilizzando la specifica
 * di Java Collections API disponibile in <code>java.util.*</code>.
 * @see lab2.Stack
 * @version 2.0 Febbraio 2001
 * @author A. Brogi      (<A HREF="mailto:brogi@di.unipi.it">brogi@di.unipi.it</A>)
 * @author A. Corradini  (<A HREF="mailto:andrea@di.unipi.it">andrea@di.unipi.it</A>)
 * @author F. Gadducci   (<A HREF="mailto:gadducci@di.unipi.it">gadducci@di.unipi.it</A>)
 * @author R. Grossi     (<A HREF="mailto:grossi@di.unipi.it">grossi@di.unipi.it</A>)
 * @author C. Montangero (<A HREF="mailto:monta@di.unipi.it">monta@di.unipi.it</A>)
 * @author F. Romani     (<A HREF="mailto:romani@di.unipi.it">romani@di.unipi.it</A>)
 * @author L. Semini     (<A HREF="mailto:semini@di.unipi.it">semini@di.unipi.it</A>)
 */
public class VectorList implements List {
    protected Vector theVector;
    protected int modCount;    // Numero di modifiche effettuate nell'insieme
    
    /**
     * Costruttore della Lista vuota.
     */
    public VectorList( ) {
	theVector = new Vector( );
    }

    /**
     * Verifica che la pila sia logicamente vuota.
     * @return  <code>true</code> se la pila &egrave; vuota;
     *          <code>false</code> altrimenti.
     */
    public boolean isEmpty( ) {
	return theVector.isEmpty( );
    }

    /**
     *  Svuota la lista.
     */
    public void clear( ) {
	modCount++;
	theVector.removeAllElements( );
    }

    /**
     * Verifica se la lista contiene l'oggetto specificato, usando il
     * metodo standard <tt>equals</tt> (l'oggetto pu&ograve; occorrere
     * pi&ugrave; volte).
     * @param x l'oggetto da cercare nella lista.
     * @return  <code>true</code> se la lista contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.
     */
    public boolean contains( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
	return theVector.contains(x);
    }

    /**
     * Inserisce un oggetto in fondo alla lista. Non sono ammessi
     * oggetti null.
     * @param x l'oggetto da inserire, <em>diverso</em> da null.
     * @return  <code>true</code> come richiesto da Java Collection API.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.
     */
    public boolean add( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
        modCount++;
        theVector.addElement( x );
        return true;
    }

    /**
     * Cancella la prima occorrenza dell'oggetto a partire dall'inizio
     * della lista, che rimane inalterata se non contiene l'oggetto.
     * Il metodo standard <tt>equals</tt> viene impiegato per trovare
     * la prima occorrenza.
     * @param x  l'oggetto da cancellare, se presente e <em>diverso</em>
     * da null.
     * @return  <code>true</code> se la lista contiene l'oggetto;
     *          <code>false</code> altrimenti.
     * @exception IllegalArgumentException se l'argomento passato
     *            &egrave; <code>null</code>.
     */
    public boolean remove( Object x ) {
	if ( x == null )
	    throw new IllegalArgumentException( );
	if ( isEmpty( )  )
	    return false;
	modCount++;
	return theVector.removeElement(x);
    }
    
    /**
     * Restituisce un iteratore sulla lista per scandire gli oggetti
     * in sequenza a partire dall'inizio della lista. Se la lista
     * &egrave; modificata con metodi diversi da quelli
     * dell'iteratore, viene generata una eccezione
     * <tt>ConcurrentModificationException</tt>. In tal caso,
     * l'iterazione termina per evitare comportamenti difficili
     * da prevedere.
     * @return una classe che realizza <code>Iterator</code> per
     * scandire gli oggetti in sequenza dall'inizio.
     * @exception lab2.Exceptions.ConcurrentModificationException se
     * avviene una modifica concorrente mediante altri metodi.
     */
    public Iterator iterator( ) {
	return new ProtectedItr( );
    }
    protected class ProtectedItr implements Iterator {
	protected int cursor;
	// Il cursore implicitamente gestito nell'iteratore
	protected boolean removed;
	protected int expectedModCount;
	// Per accorgersi di eventuali modifiche concorrenti
	
	/**
	 * Costruttore
	 */
	ProtectedItr( ) {
	    cursor = -1;
	    removed = true;
	    expectedModCount = modCount;
	    // Per il controllo delle modifiche concorrenti
	}
	
	/**
	 * Indica se ci sono ulteriori nodi da esaminare
	 * nell'iterazione corrente.
	 * @return <tt>true</tt> se ci sono ulteriori oggetti;
	 *         <tt>false</tt> altrimenti.
	 * @exception ConcurrentModificationException se avviene
	 *            una modifica concorrente degli stessi oggetti
	 *            mediante qualche altro metodo.
	 */
	public boolean hasNext( ) {
	    if ( modCount != expectedModCount )  // Modifica concorrente
		throw new ConcurrentModificationException( );
	    return ( !theVector.isEmpty( ) && (cursor < theVector.size()-1) );
	}
	
	/**
	 * Restituisce l'oggetto successivo da esaminare
	 * nell'iterazione.
	 * @return il prossimo oggetto da esaminare.
	 * @exception NoSuchElementException se non ci sono altri
	 *            oggetti da esaminare.
	 * @exception ConcurrentModificationException se avviene
	 *            una modifica concorrente degli stessi oggetti
	 *            mediante qualche altro metodo.
	 */
	public Object next( ) {
	    if ( !hasNext( ) )                   // Fuori lista
		throw new NoSuchElementException( );
	    cursor++;
	    removed = false;
	    return theVector.elementAt(cursor);
	}
	
	/**
	 * Cancella l'oggetto restituito dall'ultima chiamata al
	 * metodo <code>next</code>.  Il metodo pu&ograve; essere
	 * invocato solo una volta per ciascuna chiamata a
	 * <code>next</code>.
	 * @exception ConcurrentModificationException se avviene
	 * una modifica concorrente degli stessi oggetti
	 * mediante qualche altro metodo.
	 * @exception IllegalStateException se nessuna chiamata al
	 * metodo <tt>next</tt> &egrave; stata precedentemente
	 * effettuata, oppure una chiamata a <tt>remove</tt> &egrave;
	 * stata gi&agrave; eseguita in corrispondenza dello stessa
	 * chiamata a <tt>next</tt>.
	 */
	public void remove( ) {
	    if ( modCount++ != expectedModCount++ )  // Modifica concorrente
		throw new ConcurrentModificationException( );
	    if ( cursor == -1) // Chiamata non legale
		throw new IllegalStateException( );
	    cursor--;
	    removed=true;
	    // Il metodo remove e' introdotto in Java 1.2
	    theVector.remove(cursor);
	}
    }
}
